import os
import io
import logging
import requests
import hashlib
import tempfile
import zipfile
import dateutil.parser
from pathlib import Path
from multiformats import multihash
import binascii
from lxml import etree
from datetime import datetime, timedelta, timezone
from osgeo import ogr, osr
from pyproj import Geod
import numpy as np
import spatialist
from spatialist.raster import Raster, rasterize
from spatialist.vector import bbox, intersect, boundary, vectorize, Vector, crsConvert
import pyroSAR
from pyroSAR.ancillary import Lock, LockCollection
from pyroSAR import identify_many
from collections import defaultdict
from typing import Callable, List, TypeVar

log = logging.getLogger('cesard')

T = TypeVar('T')  # any type
K = TypeVar('K')  # key


def buffer_min_overlap(
        geom1: Vector,
        geom2: Vector,
        percent: int | float = 1,
        step: int | float | None = None
) -> None:
    """
    Buffer a rectangular geometry to a minimum overlap with a second geometry.
    The geometry is iteratively buffered until the minimum overlap is reached.
    If the overlap of the input geometries is already larger than the defined
    threshold, a copy of the original geometry is returned.

    Parameters
    ----------
    geom1:
        the geometry to be buffered
    geom2:
        the reference geometry to intersect with
    percent:
        the minimum overlap in percent of `geom1`
    step:
        the buffering step size. If None, the step size is 0.1 % of the
        average rectangle corner length.
    """
    geom1_crs = geom1.getProjection('epsg')
    geom2_crs = geom2.getProjection('epsg')
    if geom1_crs != geom2_crs:
        raise ValueError('both geometries must have the same CRS')
    geom2_area = geom2.getArea()
    ext = geom1.extent
    ext2 = ext.copy()
    if step is None:
        xdist = ext['xmax'] - ext['xmin']
        ydist = ext['ymax'] - ext['ymin']
        step = (xdist + ydist) / 2 / 1000
    buffer = 0
    overlap = 0
    while overlap <= percent:
        xbuf = buffer * step
        ybuf = buffer * step
        ext2['xmin'] = ext['xmin'] - xbuf
        ext2['xmax'] = ext['xmax'] + xbuf
        ext2['ymin'] = ext['ymin'] - ybuf
        ext2['ymax'] = ext['ymax'] + ybuf
        with bbox(ext2, geom1_crs) as geom3:
            ext3 = geom3.extent
            with intersect(geom2, geom3) as inter:
                inter_area = inter.getArea()
                overlap = inter_area / geom2_area * 100
        buffer += 1
    return bbox(ext3, geom1_crs)


def buffer_time(
        start: str,
        stop: str,
        as_datetime: bool = False,
        str_format: str = '%Y%m%dT%H%M%S',
        **kwargs
) -> tuple[str | datetime, str | datetime]:
    """
    Time range buffering

    Parameters
    ----------
    start:
        the start time date object to convert; timezone-unaware dates are interpreted as UTC.
    stop:
        the stop time date object to convert; timezone-unaware dates are interpreted as UTC.
    as_datetime:
        return datetime objects instead of strings?
    str_format:
        the output string format (ignored if `as_datetime` is True)
    kwargs
        time arguments passed to :func:`datetime.timedelta`

    Returns
    -------
        the buffered start and stop time as string or datetime object
    """
    td = timedelta(**kwargs)
    start = date_to_utc(start, as_datetime=True) - td
    stop = date_to_utc(stop, as_datetime=True) + td
    if not as_datetime:
        start = start.strftime(str_format)
        stop = stop.strftime(str_format)
    return start, stop


def check_scene_consistency(
        scenes: list[str | pyroSAR.drivers.ID]
) -> None:
    """
    Check the consistency of a scene selection.
    The following pyroSAR object attributes must be the same:
    
     - sensor
     - acquisition_mode
     - product
     - frameNumber (data take ID for Sentinel-1)
    
    Parameters
    ----------
    scenes:
        the scene selection
    
    Raises
    ------
    RuntimeError
    """
    scenes = identify_many(scenes)
    for attr in ['sensor', 'acquisition_mode', 'product', 'frameNumber']:
        values = set([getattr(x, attr) for x in scenes])
        if not len(values) == 1:
            msg = f"scene selection differs in attribute '{attr}': {values}"
            raise RuntimeError(msg)


def check_spacing(
        spacing: int | float
) -> None:
    """
    Check whether the spacing fits into the MGRS tile boundaries.

    Parameters
    ----------
    spacing: int or float
        the target pixel spacing in meters

    """
    # 109800 m is the edge length of one tile.
    # the overlap between tiles is either 9780 or 9840 m.
    overlap_edges = [9780, 9840, 109800]
    options = []
    for i in range(1, (spacing + 1) * 10):
        if all([x % (i / 10) == 0 for x in overlap_edges]):
            options.append(i / 10)
    if spacing not in options:
        raise RuntimeError(f'target spacing of {spacing} m does not align '
                           f'with MGRS tile size and overlaps.\nOptions: {options}')


def combine_polygons(
        vector: Vector | list[Vector],
        crs: int | str = 4326,
        multipolygon: bool = False,
        layer_name: str = 'combined'
) -> Vector:
    """
    Combine polygon vector objects into one.
    The output is a single vector object with the polygons either stored in
    separate features or combined into a single multipolygon geometry.

    Parameters
    ----------
    vector:
        the input vector object(s). Providing only one object only makes sense when `multipolygon=True`.
    crs:
        the target CRS. Default: EPSG:4326
    multipolygon:
        combine all polygons into one multipolygon?
        Default False: write each polygon into a separate feature.
    layer_name:
        the layer name of the output vector object.

    Returns
    -------
        the combined vector object
    """
    if not isinstance(vector, list):
        vector = [vector]
    ##############################################################################
    # check geometry types
    geometry_names = []
    field_defs = []
    for item in vector:
        field_defs.extend(item.fieldDefs)
        for feature in item.layer:
            geom = feature.GetGeometryRef()
            geometry_names.append(geom.GetGeometryName())
        item.layer.ResetReading()
    geom = None
    geometry_names = list(set(geometry_names))
    if not all(x == 'POLYGON' for x in geometry_names):
        raise RuntimeError('All geometries must be of type POLYGON')
    ##############################################################################
    vec = Vector(driver='Memory')
    srs_out = crsConvert(crs, 'osr')
    if multipolygon:
        geom_type = ogr.wkbMultiPolygon
        geom_out = [ogr.Geometry(geom_type)]
    else:
        geom_type = ogr.wkbPolygon
        geom_out = []
    fields = []
    vec.addlayer(name=layer_name, srs=srs_out, geomType=geom_type)
    for item in vector:
        fieldnames = item.fieldnames
        if item.srs.IsSame(srs_out):
            coord_trans = None
        else:
            coord_trans = osr.CoordinateTransformation(item.srs, srs_out)
        for feature in item.layer:
            geom = feature.GetGeometryRef()
            if coord_trans is not None:
                geom.Transform(coord_trans)
            if multipolygon:
                geom_out[0].AddGeometry(geom.Clone())
            else:
                fields.append({x: feature.GetField(x) for x in fieldnames})
                geom_out.append(geom.Clone())
        item.layer.ResetReading()
    geom = None
    if multipolygon:
        geom_out = geom_out[0].UnionCascaded()
        vec.addfeature(geom_out)
    else:
        for field_def in field_defs:
            if field_def.GetName() not in vec.fieldnames:
                vec.layer.CreateField(field_def)
        for i, geom in enumerate(geom_out):
            vec.addfeature(geometry=geom, fields=fields[i])
    geom_out = None
    return vec


def compute_hash(
        file_path: str,
        algorithm: str = 'sha256',
        chunk_size: int = 8192,
        multihash_encode: bool = True
) -> str:
    """
    Compute the (multi)hash of a file using the specified algorithm.

    Parameters
    ----------
    file_path:
        Path to the file.
    algorithm:
        Hash algorithm to use (default is 'sha256').
    chunk_size:
        Size of chunks to read from the file in bytes (default is 8192).
    multihash_encode:
        Encode the hash according to the
        `multihash specification <https://github.com/multiformats/multihash>`_
        (default is True)?
        The hash generated by `hashlib` will be wrapped using
        :func:`multiformats.multihash.wrap`.

    Returns
    -------
        the hexadecimal hash string of the file.

    See Also
    --------
    :mod:`hashlib`
    :mod:`multiformats.multihash`
    """
    # lookup between hashlib and multihash algorithm names; to be extended if necessary
    algo_lookup = {'sha1': 'sha1',
                   'sha256': 'sha2-256',
                   'sha512': 'sha2-512'}
    if algorithm not in algo_lookup.keys():
        raise ValueError(f'Hash algorithm must be one of {algo_lookup.keys()}')
    hash_func = getattr(hashlib, algorithm)()
    with open(file_path, 'rb') as f:
        while chunk := f.read(chunk_size):
            hash_func.update(chunk)
    if multihash_encode:
        digest = hash_func.digest()
        mh = multihash.wrap(digest, algo_lookup[algorithm])
        return mh.hex()
    else:
        return hash_func.hexdigest()


def datamask(
        measurement: str,
        dm_ras: str,
        dm_vec: str
) -> str | None:
    """
    Create data masks for a given image file.
    The created raster data mask does not contain a simple mask of nodata values.
    Rather, a boundary vector geometry containing all valid pixels is created and
    then rasterized. This boundary geometry (single polygon) is saved as `dm_vec`.
    In this case `dm_vec` is returned.
    If the input image only contains nodata values, no raster data mask is created,
    and an empty dummy vector mask is created. In this case the function will return
    `None`.


    Parameters
    ----------
    measurement:
        the binary image file
    dm_ras:
        the name of the raster data mask
    dm_vec:
        the name of the vector data mask

    Returns
    -------
        `dm_vec` if the vector data mask contains a geometry or None otherwise
    """
    
    def mask_from_array(arr, dm_vec, dm_ras, ref):
        """

        Parameters
        ----------
        arr: np.ndarray
        dm_vec: str
        dm_ras: str
        ref: spatialist.raster.Raster

        Returns
        -------
        str or None
        """
        # create a dummy vector mask if the mask only contains 0 values
        if len(arr[arr == 1]) == 0:
            Path(dm_vec).touch(exist_ok=False)
            return None
        # vectorize the raster data mask
        with vectorize(target=arr, reference=ref) as vec:
            # compute a valid data boundary geometry (vector data mask)
            with boundary(vec, expression="value=1") as bounds:
                # rasterize the vector data mask
                if not os.path.isfile(dm_ras):
                    rasterize(vectorobject=bounds, reference=ref,
                              outname=dm_ras)
                # write the vector data mask
                bounds.write(outfile=dm_vec)
        return dm_vec
    
    if os.path.isfile(dm_vec) and os.path.isfile(dm_ras):
        return None if os.path.getsize(dm_vec) == 0 else dm_vec
    
    with LockCollection([dm_vec, dm_ras]):
        if not os.path.isfile(dm_vec):
            if not os.path.isfile(dm_ras):
                with Raster(measurement) as ras:
                    arr = ras.array()
                    # create a nodata mask
                    mask = ~np.isnan(arr)
                    del arr
                    out = mask_from_array(arr=mask, dm_vec=dm_vec,
                                          dm_ras=dm_ras, ref=ras)
            else:
                # read the raster data mask
                with Raster(dm_ras) as ras:
                    mask = ras.array()
                    out = mask_from_array(arr=mask, dm_vec=dm_vec,
                                          dm_ras=dm_ras, ref=ras)
                    del mask
        else:
            if os.path.getsize(dm_vec) == 0:
                out = None
            else:
                out = dm_vec
    return out


def date_to_utc(
        date: str | datetime | None,
        as_datetime: bool = False,
        str_format: str = '%Y%m%dT%H%M%S'
) -> str | datetime | None:
    """
    convert a date object to a UTC date string or datetime object.

    Parameters
    ----------
    date:
        the date object to convert; timezone-unaware dates are interpreted as UTC.
    as_datetime:
        return a datetime object instead of a string?
    str_format:
        the output string format (ignored if `as_datetime` is True)

    Returns
    -------
        the date string or datetime object in UTC time zone
    """
    if date is None:
        return date
    elif isinstance(date, str):
        out = dateutil.parser.parse(date)
    elif isinstance(date, datetime):
        out = date
    else:
        raise TypeError('date must be a string, datetime object or None')
    if out.tzinfo is None:
        out = out.replace(tzinfo=timezone.utc)
    else:
        out = out.astimezone(timezone.utc)
    if not as_datetime:
        out = out.strftime(str_format)
    return out


def generate_unique_id(
        encoded_str: bytes,
        length: int = 4
) -> str:
    """
    
    Returns a unique product identifier as a hexadecimal string.
    The CRC-16 algorithm used to compute the unique identifier is
    CRC-CCITT (0xFFFF). The resulting CRC value is truncated to
    the number of hexadecimal characters specified by the `length`
    argument.
    
    Parameters
    ----------
    encoded_str:
        A string that should be used to generate a unique id from.
        The string needs to be encoded; e.g.: `'abc'.encode()`.
    length:
        The desired length of the output string in hexadecimal
        characters (max: 4). Values higher than 4 will be capped
        at 4, since CRC-16 only produces 16 bits.
    
    Returns
    -------
        The unique product identifier (upper-case hexadecimal string).
    """
    crc = binascii.crc_hqx(encoded_str, 0xffff)
    max_length = 4  # Max characters for 16-bit CRC
    length = max(1, min(length, max_length))  # Clamp between 1 and 4
    mask = (1 << (length * 4)) - 1  # Each hex digit = 4 bits
    p_id = f'{crc & mask:0{length}X}'
    return p_id


def get_kml() -> str:
    """
    Download the Sentinel-2 MGRS grid KML file. The target folder is ~/cesard.

    Returns
    -------
        the path to the KML file
    """
    remote = ('https://sentiwiki.copernicus.eu/__attachments/1692737/'
              'S2A_OPER_GIP_TILPAR_MPC__20151209T095117_V20150622T000000_21000101T000000_B00.zip')
    local_path = os.path.join(os.path.expanduser('~'), '.cesard')
    os.makedirs(local_path, exist_ok=True)
    local = os.path.join(local_path, os.path.basename(remote).replace('.zip', '.kml'))
    with Lock(local):
        if not os.path.isfile(local):
            log.info(f'downloading MGRS grid KML file to {local_path}')
            r = requests.get(remote)
            with zipfile.ZipFile(io.BytesIO(r.content)) as zf:
                zf.extractall(local_path)
    return local


def get_max_ext(
        geometries: list[Vector],
        buffer: float | None = None,
        crs: str | int | None = None
) -> dict[str, float]:
    """
    Gets the maximum extent from a list of geometries.
    
    Parameters
    ----------
    geometries:
        List of :class:`~spatialist.vector.Vector` geometries.
    buffer:
        The buffer in units of the geometries' CRS to add to the extent.
    crs:
        The target CRS of the extent. If None (default) the extent is
        expressed in the CRS of the input geometries.
    
    Returns
    -------
        The maximum extent of the selected :class:`~spatialist.vector.Vector`
        geometries including the chosen buffer.
    """
    max_ext = {}
    crs_list = []
    for geo in geometries:
        crs_list.append(f"EPSG:{geo.getProjection('epsg')}")
        if len(max_ext.keys()) == 0:
            max_ext = geo.extent
        else:
            ext = geo.extent
            for key in ['xmin', 'ymin']:
                if ext[key] < max_ext[key]:
                    max_ext[key] = ext[key]
            for key in ['xmax', 'ymax']:
                if ext[key] > max_ext[key]:
                    max_ext[key] = ext[key]
    crs_list = list(set(crs_list))
    if len(crs_list) > 1:
        raise RuntimeError(f'The input geometries are in different CRSs: {crs_list}')
    max_ext = dict(max_ext)
    if buffer is not None:
        max_ext['xmin'] -= buffer
        max_ext['xmax'] += buffer
        max_ext['ymin'] -= buffer
        max_ext['ymax'] += buffer
    if crs is not None:
        with bbox(coordinates=max_ext, crs=crs_list[0]) as geo:
            geo.reproject(projection=crs)
            max_ext = geo.extent
    return max_ext


def get_tmp_name(suffix: str) -> str:
    """
    Get the name of a temporary file with defined suffix.
    Files are placed in a subdirectory 'cesard' of the regular
    temporary directory so the latter is not flooded with too
    many files in case they are not properly deleted.

    Parameters
    ----------
    suffix: str
        the file suffix/extension, e.g. '.tif'

    Returns
    -------
        the temporary file name
    """
    tmpdir = os.path.join(tempfile.gettempdir(), 'cesard')
    os.makedirs(tmpdir, exist_ok=True)
    return tempfile.NamedTemporaryFile(suffix=suffix, dir=tmpdir).name


def group_by_attr(
        items: List[T],
        key_fn: Callable[[T], K]
) -> List[List[T]]:
    """
    Group items based on a key function.
    
    Parameters
    ----------
    items:
        The list of arbitrary items to group.
    key_fn:
        A function that extracts a key from each item.
    
    Returns
    -------
        A list of groups, where each group is a list of items with the same key.
    
    Example
    -------
    >>> list_in = ['abc', 'axy', 'brt', 'btk']
    >>> print(group_by_attr(list_in, lambda x: x[0]))
    [['abc', 'axy'], ['brt', 'btk']]
    
    >>> list_in = [{'a': 1}, {'a': 2}, {'a': 1}, {'a': 2}]
    >>> print(group_by_attr(list_in, lambda x: x['a']))
    [[{'a': 1}, {'a': 1}], [{'a': 2}, {'a': 2}]]
    """
    grouped = defaultdict(list)
    for item in items:
        key = key_fn(item)
        grouped[key].append(item)
    return list(grouped.values())


def group_by_time(
        scenes: list[pyroSAR.drivers.ID | str],
        time: int | float = 3
) -> list[list[pyroSAR.drivers.ID]]:
    """
    Group scenes by their acquisition time difference.

    Parameters
    ----------
    scenes:
        a list of image names
    time:
        a time difference in seconds by which to group the scenes.
        The default of 3 seconds incorporates the overlap between SLCs.

    Returns
    -------
        a list of sub-lists containing the file names of the grouped scenes
    """
    # sort images by time stamp
    scenes = identify_many(scenes, sortkey='start')
    
    if len(scenes) < 2:
        return [scenes]
    
    groups = [[scenes[0]]]
    group = groups[0]
    
    for i in range(1, len(scenes)):
        start = datetime.strptime(scenes[i].start, '%Y%m%dT%H%M%S')
        stop_pred = datetime.strptime(scenes[i - 1].stop, '%Y%m%dT%H%M%S')
        diff = abs((stop_pred - start).total_seconds())
        if diff <= time:
            group.append(scenes[i])
        else:
            groups.append([scenes[i]])
            group = groups[-1]
    return groups


def pixel_size_degrees(
        lon: float, lat: float,
        xres: float, yres: float
) -> tuple[float, float]:
    """
    Convert a pixel size from meters to degrees.

    Parameters
    ----------
    lon:
        longitude in degrees
    lat:
        latitude in degrees
    xres:
        x resolution in meters
    yres:
        y resolution in meters

    Returns
    -------
        the x and y resolution in degrees
    
    See Also
    --------
    pyproj.Geod.fwd
    """
    geod = Geod(ellps="WGS84")
    lon2, lat2, _ = geod.fwd(lon, lat, az=0, dist=yres)
    yres_deg = lat2 - lat
    lon3, lat3, _ = geod.fwd(lon, lat, az=90, dist=xres)
    xres_deg = lon3 - lon
    return xres_deg, yres_deg


def vrt_add_overviews(
        vrt: str,
        overviews: list[int],
        resampling: str = 'AVERAGE'
) -> None:
    """
    Add overviews to an existing VRT file.
    Existing overviews will be overwritten.

    Parameters
    ----------
    vrt:
        the VRT file
    overviews:
         the overview levels
    resampling:
        the overview resampling method
    """
    tree = etree.parse(vrt)
    root = tree.getroot()
    ovr = root.find('OverviewList')
    if ovr is None:
        ovr = etree.SubElement(root, 'OverviewList')
    ovr.text = ' '.join([str(x) for x in overviews])
    ovr.attrib['resampling'] = resampling.lower()
    etree.indent(root)
    tree.write(vrt, pretty_print=True, xml_declaration=False, encoding='utf-8')



import os
import re
from datetime import datetime, timezone
import numpy as np
import pandas as pd
import geopandas as gpd
from lxml import etree
from time import gmtime, strftime
from copy import deepcopy
from scipy.interpolate import RBFInterpolator
from osgeo import gdal
from spatialist.vector import bbox
from spatialist.raster import Raster, Dtype
from spatialist.auxil import gdalbuildvrt
from pyroSAR.drivers import ID

from cesard.ancillary import combine_polygons
import logging

log = logging.getLogger('cesard')


def create_vrt(
        src: str | list[str],
        dst: str,
        fun: str,
        relpaths: bool = False,
        scale: int | None = None,
        offset: float | None = None,
        dtype: str | None = None,
        args: dict[str, int | float | str] | None = None,
        options: dict | None = None,
        overviews: list[int] | None = None,
        overview_resampling: str | None = None
) -> None:
    """
    Create a GDAL VRT file executing an on-the-fly pixel function.

    Parameters
    ----------
    src:
        The input dataset(s).
    dst:
        The output dataset.
    fun:
        A `PixelFunctionType` that should be applied on the fly when opening the VRT file. The function is applied to a
        band that derives its pixel information from the source bands. A list of possible options can be found here:
        https://gdal.org/drivers/raster/vrt.html#default-pixel-functions.
        Furthermore, the option 'decibel' can be specified, which will implement a custom pixel function that uses
        Python code for decibel conversion (10*log10).
    relpaths:
        Should all `SourceFilename` XML elements with attribute `@relativeToVRT="0"` be updated to be paths relative to
        the output VRT file? Default is False.
    scale:
         The scale that should be applied when computing “real” pixel values from scaled pixel values on a raster band.
         Will be ignored if `fun='decibel'`.
    offset:
        The offset that should be applied when computing “real” pixel values from scaled pixel values on a raster band.
        Will be ignored if `fun='decibel'`.
    dtype:
        the data type of the written VRT file; default None: same data type as source data.
        data type notations of GDAL (e.g. `Float32`) and numpy (e.g. `int8`) are supported.
    args:
        arguments for `fun` passed as `PixelFunctionArguments`. Requires GDAL>=3.5 to be read.
    options:
        Additional parameters passed to :func:`osgeo.gdal.BuildVRT`.
    overviews:
        Internal overview levels to be created for each raster file.
    overview_resampling:
        Resampling method for overview levels.

    Examples
    --------
    linear gamma0 backscatter as input:

    >>> src = 's1a-iw-nrb-20220601t052704-043465-0530a1-32tpt-vh-g-lin.tif'

    decibel scaling I:
    use `log10` pixel function and additional `Scale` parameter.
    Known to display well in QGIS, but `Scale` is ignored when reading array in Python.

    >>> dst = src.replace('-lin.tif', '-log1.vrt')
    >>> create_vrt(src=src, dst=dst, fun='log10', scale=10)

    decibel scaling II:
    use custom Python pixel function. Requires additional environment variable GDAL_VRT_ENABLE_PYTHON set to YES.

    >>> dst = src.replace('-lin.tif', '-log2.vrt')
    >>> create_vrt(src=src, dst=dst, fun='decibel')

    decibel scaling III:
    use `dB` pixel function with additional `PixelFunctionArguments`. Works best but requires GDAL>=3.5.

    >>> dst = src.replace('-lin.tif', '-log3.vrt')
    >>> create_vrt(src=src, dst=dst, fun='dB', args={'fact': 10})
    """
    options = {} if options is None else options
    gdalbuildvrt(src=src, dst=dst, **options)
    tree = etree.parse(dst)
    root = tree.getroot()
    band = tree.find('VRTRasterBand')
    band.attrib['subClass'] = 'VRTDerivedRasterBand'
    
    if dtype is not None:
        band.attrib['dataType'] = Dtype(dtype).gdalstr
    
    if fun == 'decibel':
        pxfun_language = etree.SubElement(band, 'PixelFunctionLanguage')
        pxfun_language.text = 'Python'
        pxfun_type = etree.SubElement(band, 'PixelFunctionType')
        pxfun_type.text = fun
        pxfun_code = etree.SubElement(band, 'PixelFunctionCode')
        pxfun_code.text = etree.CDATA("""
    import numpy as np
    def decibel(in_ar, out_ar, xoff, yoff, xsize, ysize, raster_xsize, raster_ysize, buf_radius, gt, **kwargs):
        np.multiply(np.log10(in_ar[0], where=in_ar[0]>0.0, out=out_ar, dtype='float32'), 10.0, out=out_ar, dtype='float32')
        """)
    else:
        pixfun_type = etree.SubElement(band, 'PixelFunctionType')
        pixfun_type.text = fun
        if args is not None:
            arg = etree.SubElement(band, 'PixelFunctionArguments')
            for key, value in args.items():
                arg.attrib[key] = str(value)
        if scale is not None:
            sc = etree.SubElement(band, 'Scale')
            sc.text = str(scale)
        if offset is not None:
            off = etree.SubElement(band, 'Offset')
            off.text = str(offset)
    
    if any([overviews, overview_resampling]) is not None:
        ovr = tree.find('OverviewList')
        if ovr is None:
            ovr = etree.SubElement(root, 'OverviewList')
        if overview_resampling is not None:
            ovr.attrib['resampling'] = overview_resampling.lower()
        if overviews is not None:
            ov = str(overviews)
            for x in ['[', ']', ',']:
                ov = ov.replace(x, '')
            ovr.text = ov
    
    if relpaths:
        srcfiles = tree.xpath('//SourceFilename[@relativeToVRT="0"]')
        for srcfile in srcfiles:
            repl = os.path.relpath(srcfile.text, start=os.path.dirname(dst))
            repl = repl.replace('\\', '/')
            srcfile.text = repl
            srcfile.attrib['relativeToVRT'] = '1'
    
    etree.indent(root)
    tree.write(dst, pretty_print=True, xml_declaration=False, encoding='utf-8')


def create_rgb_vrt(
        outname: str,
        infiles: list[str],
        overviews: list[int],
        overview_resampling: str
) -> None:
    """
    Creation of the color composite GDAL VRT file.

    Parameters
    ----------
    outname:
        Full path to the output VRT file.
    infiles:
        A list of paths pointing to the linear scaled measurement backscatter files.
    overviews:
        Internal overview levels to be defined for the created VRT file.
    overview_resampling:
        Resampling method applied to overview pyramids.
    """
    
    # make sure order is right and co-polarization (VV or HH) is first
    pols = [re.search('[hv]{2}', os.path.basename(f)).group() for f in infiles]
    if pols[1] in ['vv', 'hh']:
        infiles.reverse()
        pols.reverse()
    
    # format overview levels
    ov = str(overviews)
    for x in ['[', ']', ',']:
        ov = ov.replace(x, '')
    
    # create VRT file and change its content
    gdalbuildvrt(src=infiles, dst=outname, separate=True)
    
    tree = etree.parse(outname)
    root = tree.getroot()
    srs = tree.find('SRS').text
    geotrans = tree.find('GeoTransform').text
    bands = tree.findall('VRTRasterBand')
    vrt_nodata = bands[0].find('NoDataValue').text
    complex_src = [band.find('ComplexSource') for band in bands]
    for cs in complex_src:
        cs.remove(cs.find('NODATA'))
    
    new_band = etree.SubElement(root, 'VRTRasterBand',
                                attrib={'dataType': 'Float32', 'band': '3',
                                        'subClass': 'VRTDerivedRasterBand'})
    new_band_na = etree.SubElement(new_band, 'NoDataValue')
    new_band_na.text = 'nan'
    pxfun_type = etree.SubElement(new_band, 'PixelFunctionType')
    pxfun_type.text = 'mul'
    for cs in complex_src:
        new_band.append(deepcopy(cs))
    
    src = new_band.findall('ComplexSource')[1]
    fname = src.find('SourceFilename')
    fname_old = fname.text
    src_attr = src.find('SourceProperties').attrib
    fname.text = etree.CDATA("""
    <VRTDataset rasterXSize="{rasterxsize}" rasterYSize="{rasterysize}">
        <SRS dataAxisToSRSAxisMapping="1,2">{srs}</SRS>
        <GeoTransform>{geotrans}</GeoTransform>
        <VRTRasterBand dataType="{dtype}" band="1" subClass="VRTDerivedRasterBand">
            <NoDataValue>{vrt_nodata}</NoDataValue>
            <PixelFunctionType>{px_fun}</PixelFunctionType>
            <ComplexSource>
              <SourceFilename relativeToVRT="1">{fname}</SourceFilename>
              <SourceBand>1</SourceBand>
              <SourceProperties RasterXSize="{rasterxsize}" RasterYSize="{rasterysize}" DataType="{dtype}" BlockXSize="{blockxsize}" BlockYSize="{blockysize}"/>
              <SrcRect xOff="0" yOff="0" xSize="{rasterxsize}" ySize="{rasterysize}"/>
              <DstRect xOff="0" yOff="0" xSize="{rasterxsize}" ySize="{rasterysize}"/>
            </ComplexSource>
        </VRTRasterBand>
        <OverviewList resampling="{ov_resampling}">{ov}</OverviewList>
    </VRTDataset>
    """.format(rasterxsize=src_attr['RasterXSize'], rasterysize=src_attr['RasterYSize'], srs=srs, geotrans=geotrans,
               dtype=src_attr['DataType'], px_fun='inv', fname=fname_old, vrt_nodata=vrt_nodata,
               blockxsize=src_attr['BlockXSize'], blockysize=src_attr['BlockYSize'],
               ov_resampling=overview_resampling.lower(), ov=ov))
    
    bands = tree.findall('VRTRasterBand')
    for band, col in zip(bands, ['Red', 'Green', 'Blue']):
        color = etree.Element('ColorInterp')
        color.text = col
        band.insert(0, color)
    
    ovr = etree.SubElement(root, 'OverviewList', attrib={'resampling': overview_resampling.lower()})
    ovr.text = ov
    
    etree.indent(root)
    tree.write(outname, pretty_print=True, xml_declaration=False, encoding='utf-8')


def calc_product_start_stop(
        src_ids: list[ID],
        extent: dict[str, int | float],
        epsg: int
) -> tuple[datetime, datetime]:
    """
    Calculates the start and stop times of the ARD product.
    The geolocation grid points including their azimuth time information are
    extracted first from the metadata of each source product.
    These grid points are then used to interpolate the azimuth time for the
    coordinates of the MGRS tile extent. The lowest and highest interpolated
    value are returned as product acquisition start and stop times of the
    ARD product.

    Parameters
    ----------
    src_ids:
        List of :class:`~pyroSAR.drivers.ID` objects of all source products
        that overlap with the current MGRS tile.
    extent:
        Spatial extent of the MGRS tile, derived from a
        :class:`~spatialist.vector.Vector` object.
    epsg:
        The coordinate reference system of the extent as an EPSG code.

    Returns
    -------
        Start and stop time of the ARD product in UTC.
    
    See Also
    --------
    pyroSAR.drivers.SAFE.geo_grid
    scipy.interpolate.RBFInterpolator
    """
    with bbox(extent, epsg) as tile_geom:
        tile_geom.reproject(4326)
        scene_geoms = [x.geometry() for x in src_ids]
        with combine_polygons(scene_geoms) as scene_geom:
            intersection = gpd.overlay(df1=tile_geom.to_geopandas(),
                                       df2=scene_geom.to_geopandas(),
                                       how='intersection')
            tile_geom_pts = intersection.get_coordinates().to_numpy()
        scene_geoms = None
    
    # combine geo grid of all scenes into one
    gdfs = []
    for src_id in src_ids:
        with src_id.geo_grid() as vec:
            gdfs.append(vec.to_geopandas())
    gdf = pd.concat(gdfs, ignore_index=True)
    
    # remove duplicate points
    gdf["xy"] = gdf.geometry.apply(lambda p: (p.x, p.y))
    gdf = gdf.drop_duplicates(subset="xy").copy()
    gdf.drop(columns="xy", inplace=True)
    
    # get grid point coordinates and numerical time stamps for interpolation
    gdf['timestamp'] = gdf['azimuthTime'].astype(np.int64) / 10 ** 9
    gridpts = gdf.get_coordinates().to_numpy()
    az_time = gdf['timestamp'].values
    
    # perform interpolation
    rbf = RBFInterpolator(y=gridpts, d=az_time)
    interpolated = rbf(tile_geom_pts)
    
    # check interpolation validity
    if np.isnan(interpolated).any():
        raise RuntimeError('The interpolated array contains NaN values.')
    
    # Make sure the interpolated values do not exceed the actual values.
    # This might happen when the source product geometries are slightly
    # larger than the geo grid extent.
    out = [max(min(interpolated), min(gdf['timestamp'])),
           min(max(interpolated), max(gdf['timestamp']))]
    
    # double-check that values are plausible
    if out[0] < min(gdf['timestamp']) or out[1] > max(gdf['timestamp']):
        raise RuntimeError('The interpolated values exceed the input range.')
    if out[0] >= out[1]:
        raise RuntimeError('The determined acquisition start is larger '
                           'than or equal to the acquisition end.')
    
    return (datetime.fromtimestamp(out[0], tz=timezone.utc),
            datetime.fromtimestamp(out[1], tz=timezone.utc))


def create_data_mask(
        outname: str,
        datasets: list[dict],
        extent: dict[str, int | float],
        epsg: int,
        driver: str,
        creation_opt: list[str],
        overviews: list[int],
        overview_resampling: str,
        dst_nodata: int | str,
        product_type: str,
        lsm_encoding: dict[str, int],
        wbm: str | None = None
) -> None:
    """
    Creation of the Data Mask image.
    
    Parameters
    ----------
    outname:
        Full path to the output data mask file.
    datasets:
        List of processed output files that match the source scenes and overlap
        with the current MGRS tile. An error will be thrown if not all datasets
        contain a key `datamask`. The function will return without an error if
        not all datasets contain a key `dm`.
    extent:
        Spatial extent of the MGRS tile, derived from a
        :class:`~spatialist.vector.Vector` object.
    epsg:
        The coordinate reference system as an EPSG code.
    driver:
        GDAL driver to use for raster file creation.
    creation_opt:
        GDAL creation options to use for raster file creation. Should match
        specified GDAL driver.
    overviews:
        Internal overview levels to be created for each raster file.
    overview_resampling:
        Resampling method for overview levels.
    dst_nodata:
        Nodata value to write to the output raster.
    product_type:
        The type of ARD product that is being created. Either 'NRB' or 'ORB'.
    lsm_encoding:
        a dictionary containing the layover shadow mask encoding.
    wbm:
        Path to a water body mask file with the dimensions of an MGRS tile.
        Optional if `product_type='NRB', mandatory if `product_type='ORB'`.
    """
    measurement_keys = [x for x in datasets[0].keys() if re.search('[gs]-lin', x)]
    measurement = [scene[measurement_keys[0]] for scene in datasets]
    datamask = [scene['datamask'] for scene in datasets]
    ls = []
    for scene in datasets:
        if 'dm' in scene:
            ls.append(scene['dm'])
        else:
            return  # do not create a data mask if not all scenes have a layover-shadow mask
    
    dm_bands = ['not layover, nor shadow',
                'layover',
                'shadow',
                'ocean',
                'lakes',
                'rivers']
    
    if product_type == 'ORB':
        if wbm is None:
            raise RuntimeError('Water body mask is required for ORB products')
    
    tile_bounds = [extent['xmin'], extent['ymin'], extent['xmax'], extent['ymax']]
    
    vrt_ls = '/vsimem/' + os.path.dirname(outname) + 'ls.vrt'
    vrt_valid = '/vsimem/' + os.path.dirname(outname) + 'valid.vrt'
    vrt_measurement = '/vsimem/' + os.path.dirname(outname) + 'measurement.vrt'
    gdalbuildvrt(src=ls, dst=vrt_ls, outputBounds=tile_bounds, void=False)
    gdalbuildvrt(src=datamask, dst=vrt_valid, outputBounds=tile_bounds, void=False)
    gdalbuildvrt(src=measurement, dst=vrt_measurement, outputBounds=tile_bounds, void=False)
    
    with Raster(vrt_ls) as ras_ls:
        with bbox(extent, crs=epsg) as tile_vec:
            ras_ls_res = ras_ls.res
            rows = ras_ls.rows
            cols = ras_ls.cols
            geotrans = ras_ls.raster.GetGeoTransform()
            proj = ras_ls.raster.GetProjection()
            arr_dm = ras_ls.array()
            
            # Get Water Body Mask
            if wbm is not None:
                with Raster(wbm) as ras_wbm:
                    ras_wbm_cols = ras_wbm.cols
                    cols_ratio = ras_wbm_cols / cols
                    if cols_ratio > 1:
                        # create low resolution VRT
                        wbm_lowres = wbm.replace('.tif', f'_{ras_ls_res[0]}m.vrt')
                        if not os.path.isfile(wbm_lowres):
                            options = {'xRes': ras_ls_res[0], 'yRes': ras_ls_res[1],
                                       'resampleAlg': 'mode'}
                            gdalbuildvrt(src=wbm, dst=wbm_lowres, **options)
                        with Raster(wbm_lowres) as ras_wbm_lowres:
                            arr_wbm = ras_wbm_lowres.array()
                    else:
                        arr_wbm = ras_wbm.array()
            else:
                del dm_bands[3:]
            
            c = lsm_encoding['not layover, not shadow']
            l = lsm_encoding['layover']
            s = lsm_encoding['shadow']
            ls = lsm_encoding['layover in shadow']
            n = lsm_encoding['nodata']
            
            # Extend the shadow class of the data mask with nodata values
            # from backscatter data and create final array
            with Raster(vrt_valid)[tile_vec] as ras_valid:
                with Raster(vrt_measurement)[tile_vec] as ras_measurement:
                    arr_valid = ras_valid.array()
                    arr_measurement = ras_measurement.array()
                    
                    arr_dm[~np.isfinite(arr_dm)] = n
                    arr_dm = np.where(((arr_valid == 1) & (np.isnan(arr_measurement))),
                                      s, arr_dm)
                    arr_dm[arr_valid != 1] = n
                    del arr_measurement
                    del arr_valid
        
        outname_tmp = '/vsimem/' + os.path.basename(outname) + '.vrt'
        gdriver = gdal.GetDriverByName('GTiff')
        ds_tmp = gdriver.Create(outname_tmp, rows, cols, len(dm_bands), gdal.GDT_Byte,
                                options=['ALPHA=UNSPECIFIED', 'PHOTOMETRIC=MINISWHITE'])
        gdriver = None
        ds_tmp.SetGeoTransform(geotrans)
        ds_tmp.SetProjection(proj)
        
        for i, name in enumerate(dm_bands):
            band = ds_tmp.GetRasterBand(i + 1)
            
            # not layover, nor shadow
            if i == 0:
                arr = arr_dm == c
            # layover | layover in shadow
            elif i == 1:
                arr = (arr_dm == l) | (arr_dm == ls)
            # shadow | layover in shadow
            elif i == 2:
                arr = (arr_dm == s) | (arr_dm == ls)
            # ocean
            elif i == 3:
                arr = arr_wbm == 1
            # lakes
            elif i == 4:
                arr = arr_wbm == 2
            # rivers
            elif i == 5:
                arr = arr_wbm == 3
            else:
                raise ValueError(f'unknown array value: {i}')
            
            arr = arr.astype('uint8')
            arr[arr_dm == n] = dst_nodata
            band.WriteArray(arr)
            band.SetNoDataValue(dst_nodata)
            band.SetDescription(name)
            band.FlushCache()
            band = None
            del arr
        
        ds_tmp.SetMetadataItem('TIFFTAG_DATETIME', strftime('%Y:%m:%d %H:%M:%S', gmtime()))
        ds_tmp.BuildOverviews(overview_resampling, overviews)
        outDataset_cog = gdal.GetDriverByName(driver).CreateCopy(outname, ds_tmp, strict=1, options=creation_opt)
        outDataset_cog = None
        ds_tmp = None
        tile_vec = None


def create_acq_id_image(
        outname: str,
        ref_tif: str,
        datasets: list[dict],
        src_ids: list[ID],
        extent: dict[str, int | float],
        epsg: int,
        driver: str,
        creation_opt: list[str],
        overviews: list[int],
        dst_nodata: int | str
) -> None:
    """
    Creation of the Acquisition ID image.

    Parameters
    ----------
    outname:
        Full path to the output data mask file.
    ref_tif:
        Full path to any GeoTIFF file of the ARD product.
    datasets:
        List of processed output files that match the source SLC scenes and overlap with the current MGRS tile.
    src_ids:
        List of :class:`~pyroSAR.drivers.ID` objects of all source SLC scenes that overlap with the current MGRS tile.
    extent:
        Spatial extent of the MGRS tile, derived from a :class:`~spatialist.vector.Vector` object.
    epsg:
        The CRS used for the ARD product; provided as an EPSG code.
    driver:
        GDAL driver to use for raster file creation.
    creation_opt:
        GDAL creation options to use for raster file creation. Should match specified GDAL driver.
    overviews:
        Internal overview levels to be created for each raster file.
    dst_nodata:
        Nodata value to write to the output raster.
    """
    src_scenes = [sid.scene for sid in src_ids]
    
    tile_bounds = [extent['xmin'], extent['ymin'], extent['xmax'], extent['ymax']]
    
    arr_list = []
    for dataset in datasets:
        vrt_valid = '/vsimem/' + os.path.dirname(outname) + 'mosaic.vrt'
        gdalbuildvrt(src=dataset['datamask'], dst=vrt_valid, outputBounds=tile_bounds, void=False)
        with bbox(extent, crs=epsg) as tile_vec:
            with Raster(vrt_valid)[tile_vec] as vrt_ras:
                vrt_arr = vrt_ras.array()
                arr_list.append(vrt_arr)
                del vrt_arr
            tile_vec = None
    
    src_scenes_clean = [os.path.basename(src).replace('.zip', '').replace('.SAFE', '') for src in src_scenes]
    tag = '{{"{src1}": 1}}'.format(src1=src_scenes_clean[0])
    out_arr = np.full(arr_list[0].shape, dst_nodata)
    out_arr[arr_list[0] == 1] = 1
    if len(arr_list) == 2:
        out_arr[arr_list[1] == 1] = 2
        tag = '{{"{src1}": 1, "{src2}": 2}}'.format(src1=src_scenes_clean[0], src2=src_scenes_clean[1])
    
    creation_opt.append('TIFFTAG_IMAGEDESCRIPTION={}'.format(tag))
    with Raster(ref_tif) as ref_ras:
        ref_ras.write(outname, format=driver, array=out_arr.astype('uint8'), nodata=dst_nodata, overwrite=True,
                      overviews=overviews, options=creation_opt)



import os
import re
import tempfile
import itertools
from getpass import getpass
from pyroSAR.drivers import ID
from pyroSAR.auxdata import dem_autoload, dem_create
from pyroSAR.ancillary import Lock
import cesard.tile_extraction as tile_ex
from cesard.ancillary import (get_max_ext, get_tmp_name,
                              pixel_size_degrees, vrt_add_overviews)
from spatialist.vector import bbox, intersect, Vector
from typing import Literal
import logging

log = logging.getLogger('cesard')


def authenticate(
        dem_type: str,
        username: str | None = None,
        password: str | None = None
) -> tuple[str | None, str | None]:
    """
    Query the username and password. If None, environment variables DEM_USER and DEM_PASS are read.
    If they are also None, the user is queried interactively.
    
    Parameters
    ----------
    dem_type:
        the DEM type. Needed for determining whether authentication is needed.
    username:
        The username for accessing the DEM tiles. If None and authentication is required
        for the selected DEM type, the environment variable 'DEM_USER' is read.
        If this is not set, the user is prompted interactively to provide credentials.
    password:
        The password for accessing the DEM tiles.
        If None: same behavior as for username but with env. variable 'DEM_PASS'.

    Returns
    -------
        the username and password
    """
    dems_auth = ['Copernicus 10m EEA DEM',
                 'Copernicus 30m Global DEM II']
    if dem_type not in dems_auth:
        return None, None
    
    if username is None:
        username = os.getenv('DEM_USER')
    if password is None:
        password = os.getenv('DEM_PASS')
    if username is None:
        username = input('Please enter your DEM access username:')
    if password is None:
        password = getpass('Please enter your DEM access password:')
    return username, password


def mosaic(
        geometry: Vector,
        dem_type: str,
        outname: str,
        tr: None | tuple[int | float, int | float] = None,
        username: str | None = None,
        password: str | None = None,
        threads: int = 4
) -> None:
    """
    Create a new scene-specific DEM mosaic GeoTIFF file.
    Makes use of :func:`pyroSAR.auxdata.dem_autoload` and
    :func:`pyroSAR.auxdata.dem_create`.
    
    Parameters
    ----------
    geometry:
        The geometry to be covered by the mosaic. The geometry's CRS is
        used as target CRS.
    dem_type:
        The DEM type.
    outname:
        The name of the mosaic.
    tr:
        the target resolution as (xres, yres) in units of the target CRS.
    username:
        The username for accessing the DEM tiles. If None and authentication
        is required for the selected DEM type, the environment variable
        'DEM_USER' is read. If this is not set, the user is prompted
        interactively to provide credentials.
    password:
        The password for accessing the DEM tiles.
        If None: same behavior as for username but with env. variable
        'DEM_PASS'.
    threads:
        The number of threads to pass to :func:`pyroSAR.auxdata.dem_create`.
    """
    epsg = geometry.getProjection('epsg')
    ext = geometry.extent
    if not os.path.isfile(outname):
        username, password = authenticate(dem_type=dem_type,
                                          username=username,
                                          password=password)
        if dem_type == 'GETASSE30':
            geoid_convert = False
        else:
            geoid_convert = True
        geoid = 'EGM2008'
        vrt = outname.replace('.tif', '.vrt')
        if epsg != 4326:
            geometry = geometry.clone()
            geometry.reproject(4326)
        dem_autoload([geometry], demType=dem_type,
                     vrt=vrt, buffer=0.01, product='dem',
                     username=username, password=password)
        bounds = [ext['xmin'], ext['ymin'], ext['xmax'], ext['ymax']]
        dem_create(src=vrt, dst=outname, pbar=False, tr=tr,
                   geoid_convert=geoid_convert, geoid=geoid,
                   threads=threads, nodata=-32767, t_srs=epsg,
                   outputBounds=bounds)
        os.remove(vrt)
        if epsg != 4326:
            geometry = None


def prepare(
        scene: ID,
        dem_type: str,
        mode: Literal['single-4326', 'multi-UTM'],
        dir_out: str,
        tr: None | tuple[int | float, int | float] = None,
        username: str | None = None,
        password: str | None = None
) -> list[str]:
    """
    Prepare DEM files for SAR processing.

    Parameters
    ----------
    scene:
        the SAR product
    dem_type:
        the DEM type
    mode:
        the DEM preparation mode (depends on the requirements of the used SAR processor)
    dir_out:
        the destination directory
    tr:
        the target resolution in meters as (x, y).
    username:
        The username for accessing the DEM tiles. If None and authentication is required
        for the selected DEM type, the environment variable 'DEM_USER' is read.
        If this is not set, the user is prompted interactively to provide credentials.
    password:
        The password for accessing the DEM tiles.
        If None: same behavior as for username but with env. variable 'DEM_PASS'.

    Returns
    -------
        the names of the newly created DEM files.
    """
    dem_type_lookup = {'Copernicus 10m EEA DEM': 'EEA10',
                       'Copernicus 30m Global DEM II': 'GLO30II',
                       'Copernicus 30m Global DEM': 'GLO30',
                       'GETASSE30': 'GETASSE30'}
    dem_type_short = dem_type_lookup[dem_type]
    if mode == 'single-4326':
        fname_base_dem = f'DEM_{dem_type_short}_4326.tif'
        fname_dem = os.path.join(dir_out, fname_base_dem)
        with Lock(fname_dem):
            if not os.path.isfile(fname_dem):
                log.info('creating scene-specific DEM mosaic in EPSG:4326')
                if tr is not None:
                    with scene.geometry() as vec:
                        ext = vec.extent
                    tr = pixel_size_degrees(lon=abs(ext['xmax'] - ext['xmin']) / 2,
                                            lat=abs(ext['ymax'] - ext['ymin']) / 2,
                                            xres=tr[0], yres=tr[1])
                with scene.bbox(buffer=0.002) as geom:
                    mosaic(geometry=geom, outname=fname_dem,
                           dem_type=dem_type, tr=tr,
                           username=username, password=password)
            else:
                log.info(f'found scene-specific DEM mosaic: {fname_dem}')
    elif mode == 'multi-UTM':
        aois = tile_ex.aoi_from_scene(scene=scene, multi=True)
        fname_dem = []
        for aoi in aois:
            ext = aoi['extent_utm']
            epsg = aoi['epsg']
            # reduce DEM extent to image extent but keep coordinates a multiple of 60
            # to fit into the MGRS tile boundaries
            with scene.bbox() as vec_scene:
                vec_scene.reproject(epsg)
                with bbox(coordinates=ext, crs=epsg) as vec_tiles:
                    with intersect(vec_scene, vec_tiles) as inter:
                        ext = inter.extent
                        ext['xmin'] = ext['xmin'] // 60 * 60
                        ext['ymin'] = ext['ymin'] // 60 * 60 + 60
                        ext['xmax'] = ext['xmax'] // 60 * 60
                        ext['ymax'] = ext['ymax'] // 60 * 60 + 60
            fname_base_dem = f'DEM_{dem_type_short}_{epsg}.tif'
            fname_dem_tmp = os.path.join(dir_out, fname_base_dem)
            fname_dem.append(fname_dem_tmp)
            with Lock(fname_dem_tmp):
                if not os.path.isfile(fname_dem_tmp):
                    log.info(f'creating scene-specific DEM mosaic in EPSG:{epsg}')
                    with bbox(coordinates=ext, crs=epsg, buffer=240) as geom:
                        mosaic(geometry=geom, outname=fname_dem_tmp,
                               dem_type=dem_type, tr=tr,
                               username=username, password=password)
                else:
                    log.info(f'found scene-specific DEM mosaic: {fname_dem_tmp}')
    else:
        raise ValueError('mode must be one of "single-4326" or "multi-UTM"')
    return fname_dem


def retile(
        vector: Vector,
        dem_type: str,
        dem_dir: str | None,
        wbm_dir: str | None,
        dem_strict: bool = True,
        tilenames: list[str] | None = None,
        threads: int | None = None,
        username: str | None = None,
        password: str | None = None,
        lock_timeout: int = 1200
) -> None:
    """
    Download and retile DEM and WBM tiles to MGRS.
    Including re-projection and vertical datum conversion.

    Parameters
    ----------
    vector:
        The vector object for which to prepare the DEM and WBM tiles.
        CRS must be EPSG:4236.
    dem_type:
        The DEM type.
    dem_dir:
        The DEM target directory. DEM preparation can be skipped if set to None.
    wbm_dir:
        The WBM target directory. WBM preparation can be skipped if set to None
    dem_strict:
        strictly only create DEM tiles in the native CRS of the MGRS tile or
        also allow reprojection to ensure full coverage of the vector object in every CRS.
    tilenames:
        an optional list of MGRS tile names. Default None: process all overalapping tiles.
    threads:
        The number of threads to pass to :func:`pyroSAR.auxdata.dem_create`.
        Default `None`: use the value of `GDAL_NUM_THREADS` without modification.
    username:
        The username for accessing the DEM tiles. If None and authentication is required
        for the selected DEM type, the environment variable 'DEM_USER' is read.
        If this is not set, the user is prompted interactively to provide credentials.
    password:
        The password for accessing the DEM tiles.
        If None: same behavior as for username but with env. variable 'DEM_PASS'.
    lock_timeout:
        how long to wait to acquire a lock on created files?

    Examples
    --------
    >>> from cesard import dem
    >>> from spatialist import bbox
    >>> ext = {'xmin': 12, 'xmax': 13, 'ymin': 50, 'ymax': 51}
    # strictly only create overlapping DEM tiles in their native CRS.
    # Will create tiles 32UQA, 32UQB, 33UUR and 33UUS.
    >>> with bbox(coordinates=ext, crs=4326) as vec:
    >>>     dem.retile(vector=vec, dem_type='Copernicus 30m Global DEM',
    >>>                dem_dir='DEM', wbm_dir=None, dem_strict=True,
    >>>                threads=4)
    # Process all overlapping DEM tiles to each CRS.
    # Will additionally create tiles 32UQA_32633, 32UQB_32633, 33UUR_32632 and 33UUS_32632.
    >>> with bbox(coordinates=ext, crs=4326) as vec:
    >>>     dem.retile(vector=vec, dem_type='Copernicus 30m Global DEM',
    >>>                dem_dir='DEM', wbm_dir=None, dem_strict=False,
    >>>                threads=4)

    See Also
    --------
    cesard.tile_extraction.tile_from_aoi
    """
    if dem_type == 'GETASSE30':
        geoid_convert = False
    else:
        geoid_convert = True
    geoid = 'EGM2008'  # applies to all Copernicus DEM options
    
    tr = 10  # target resolution. Lower resolutions can be created virtually using VRTs.
    # additional creation options for gdalwarp
    create_options = ['COMPRESS=LERC_ZSTD', 'MAX_Z_ERROR=0']
    
    # DEM options with WBMs
    wbm_dems = ['Copernicus 10m EEA DEM',
                'Copernicus 30m Global DEM',
                'Copernicus 30m Global DEM II']
    if wbm_dir is not None and dem_type in wbm_dems:
        wbm_dir = os.path.join(wbm_dir, dem_type)
        os.makedirs(wbm_dir, exist_ok=True)
    else:
        wbm_dir = None
    if dem_dir is not None:
        dem_dir = os.path.join(dem_dir, dem_type)
        os.makedirs(dem_dir, exist_ok=True)
    if wbm_dir is None and dem_dir is None:
        return
    # get the geometries of all tiles overlapping with the AOI
    tiles = tile_ex.tile_from_aoi(vector=vector,
                                  return_geometries=True,
                                  tilenames=tilenames)
    # group the returned tiles by CRS and process them separately
    for epsg, group in itertools.groupby(tiles, lambda x: x.getProjection('epsg')):
        vectors = list(group)
        
        # In case the DEM tiles are to be prepared as well, create a new list of tiles.
        # This new list contains all tiles covering the AOI but all re-projected to the
        # current CRS. This way, a DEM mosaic can be created from prepared tiles in any
        # UTM zone covering the AOI while fully covering it. This was needed for processing
        # full SAR scenes to different UTM zones. In the current workflow this in no longer
        # used.
        if dem_dir is not None and not dem_strict:
            vectors = tile_ex.tile_from_aoi(
                vector=vector.bbox(),
                epsg=epsg,
                strict=False,
                return_geometries=True,
                tilenames=tilenames)
        
        # Get the bounding box of the tile vector objects and use this from here on
        ext = get_max_ext(geometries=vectors, buffer=200)
        with bbox(coordinates=ext, crs=epsg) as box:
            box.reproject(4326)
            ext_4326 = box.extent
        
        if dem_dir is not None:
            dem_names_base = ['{}_DEM.tif'.format(tile.mgrs) for tile in vectors]
            dem_names = [os.path.join(dem_dir, x) for x in dem_names_base]
            dem_target = [(tile, name) for tile, name in zip(vectors, dem_names)
                          if not os.path.isfile(name)]
            fname_dem_tmp = tempfile.NamedTemporaryFile(suffix='.vrt', dir=dem_dir).name
        else:
            dem_target = dict()
            fname_dem_tmp = None
        if wbm_dir is not None:
            # exclude the reprojected tiles from the list of WBM tiles
            tiles_wbm = [x for x in vectors if not re.search('_[0-9]*', x.mgrs)]
            wbm_names_base = ['{}_WBM.tif'.format(tile.mgrs) for tile in tiles_wbm]
            wbm_names = [os.path.join(wbm_dir, x) for x in wbm_names_base]
            wbm_target = [(tile, name) for tile, name in zip(tiles_wbm, wbm_names)
                          if not os.path.isfile(name)]
            fname_wbm_tmp = tempfile.NamedTemporaryFile(suffix='.vrt', dir=wbm_dir).name
        else:
            wbm_target = dict()
            fname_wbm_tmp = None
        
        # stop if no files need to be created
        if len(dem_target) == 0 and len(wbm_target) == 0:
            continue
        ###############################################
        # DEM/WBM download and VRT mosaic creation
        
        # get download authentication if either WBM or DEM VRTs will be created
        c_wbm = fname_wbm_tmp is not None and not os.path.isfile(fname_wbm_tmp)
        c_dem = fname_dem_tmp is not None and not os.path.isfile(fname_dem_tmp)
        if c_wbm or c_dem:
            username, password = authenticate(dem_type=dem_type,
                                              username=username,
                                              password=password)
        
        # download WBM tiles and combine them in a VRT mosaic
        if c_wbm:
            with Lock(fname_wbm_tmp, timeout=lock_timeout):
                if not os.path.isfile(fname_wbm_tmp):
                    with bbox(coordinates=ext_4326, crs=4326) as vec:
                        dem_autoload(geometries=[vec], demType=dem_type,
                                     vrt=fname_wbm_tmp, product='wbm',
                                     username=username, password=password,
                                     crop=False, lock_timeout=lock_timeout)
        # download DEM tiles and combine them in a VRT mosaic
        if c_dem:
            with Lock(fname_dem_tmp, timeout=lock_timeout):
                if not os.path.isfile(fname_dem_tmp):
                    with bbox(coordinates=ext_4326, crs=4326) as vec:
                        dem_autoload(geometries=[vec], demType=dem_type,
                                     vrt=fname_dem_tmp, product='dem',
                                     username=username, password=password,
                                     crop=False, lock_timeout=lock_timeout)
        ###############################################
        # create final DEM tiles
        if len(dem_target) > 0:
            tiles = [x[0].mgrs for x in dem_target]
            log.info(f'creating DEM MGRS tiles: {tiles}')
        for tile, filename in dem_target:
            ext = tile.extent
            bounds = [ext['xmin'], ext['ymin'],
                      ext['xmax'], ext['ymax']]
            with Lock(filename, timeout=lock_timeout):
                if not os.path.isfile(filename):
                    dem_create(src=fname_dem_tmp, dst=filename,
                               t_srs=epsg, tr=(tr, tr), pbar=False,
                               geoid_convert=geoid_convert, geoid=geoid,
                               outputBounds=bounds, threads=threads,
                               nodata=-32767, creationOptions=create_options)
        if fname_dem_tmp is not None:
            os.remove(fname_dem_tmp)
        ###############################################
        # create final WBM tiles
        if len(wbm_target) > 0:
            tiles = [x[0].mgrs for x in wbm_target]
            log.info(f'creating WBM MGRS tiles: {tiles}')
        for tile, filename in wbm_target:
            ext = tile.extent
            bounds = [ext['xmin'], ext['ymin'],
                      ext['xmax'], ext['ymax']]
            with Lock(filename):
                if not os.path.isfile(filename):
                    dem_create(src=fname_wbm_tmp, dst=filename,
                               t_srs=epsg, tr=(tr, tr),
                               resampleAlg='mode', pbar=False,
                               outputBounds=bounds, threads=threads,
                               creationOptions=create_options)
        if fname_wbm_tmp is not None:
            os.remove(fname_wbm_tmp)


def to_mgrs(
        tile: str,
        dst: str,
        dem_type: str,
        overviews: list[int],
        tr: tuple[int | float, int | float],
        format: str = 'COG',
        create_options: list[str] | None = None,
        threads: int | None = None,
        pbar: bool = False
) -> None:
    """
    Create an MGRS-tiled DEM file.
    
    Parameters
    ----------
    tile:
        the MGRS tile ID
    dst:
        the destination file name
    dem_type:
        The DEM type.
    overviews:
        The overview levels
    tr:
        the target resolution as (x, y)
    format:
        the output file format
    create_options:
        additional creation options to be passed to :func:`spatialist.auxil.gdalwarp`.
    threads:
        The number of threads to pass to :func:`pyroSAR.auxdata.dem_create`.
        Default `None`: use the value of `GDAL_NUM_THREADS` without modification.
    pbar:
        add a progress bar?
    """
    if dem_type == 'GETASSE30':
        geoid_convert = False
    else:
        geoid_convert = True
    geoid = 'EGM2008'  # applies to all Copernicus DEM options
    with tile_ex.aoi_from_tile(tile=tile) as vec:
        ext = vec.extent
        epsg = vec.getProjection('epsg')
    bounds = [ext['xmin'], ext['ymin'], ext['xmax'], ext['ymax']]
    buffer = 200
    ext['xmin'] -= buffer
    ext['ymin'] -= buffer
    ext['xmax'] += buffer
    ext['ymax'] += buffer
    vrt = get_tmp_name(suffix='.vrt')
    with bbox(coordinates=ext, crs=epsg) as vec:
        vec.reproject(4326)
        dem_autoload(geometries=[vec], demType=dem_type, vrt=vrt)
    vrt_add_overviews(vrt=vrt, overviews=overviews)
    dem_create(src=vrt, dst=dst, t_srs=epsg, tr=tr,
               geoid_convert=geoid_convert, geoid=geoid, pbar=pbar,
               outputBounds=bounds, threads=threads, format=format,
               creationOptions=create_options)
    os.remove(vrt)
